/*
*************************************************************
Stewart MacLeod			SQL Server 6.X	

Debit Savings Credit Checking transaction using a stored procedure
	
	* triggers for referential integrity
	* stored procedures with error handling
	* rules to restrict values in a column
	* defaults to prevent / replace NULL values

*************************************************************
*/


/*
*********************************************************************
	Drop the example tables if they exist
*********************************************************************
*/

if exists (select * from sysobjects where name = 'savings' and type = 'U')
begin
	drop table savings
end
go

if exists (select * from sysobjects where name = 'checking' and type = 'U')
begin
	drop table checking
end
go


if exists (select * from sysobjects where name = 'customer' and type = 'U')
begin
	drop table customer
end
go


if exists (select * from sysobjects where name = 'do_transfer' )
begin
	drop proc do_transfer
end
go

if exists (select * from sysobjects where name = 'checking_insert' )
begin
	drop trigger checking_insert
end
go

if exists (select * from sysobjects where name = 'checking_update' )
begin
	drop trigger checking_update
end
go


if exists (select * from sysobjects where name = 'customer_delete' )
begin
	drop trigger customer_delete
end
go


/*
*********************************************************************
	Create the customer table
*********************************************************************
*/

create table customer
(
	custid	integer not null identity( 1,1),
	name	character(30) not null,
	phone	character(14) not null
	
)
go

/*
*********************************************************************
	Create the savings table
*********************************************************************
*/


create table savings
(
	custid	integer not null,
	balance	decimal(10,2) not null	
)
go

/*
*********************************************************************
	Create the checking table
*********************************************************************
*/

create table checking
(
	custid	integer not null,
	balance	decimal(10,2) not null	
)
go

/*
*********************************************************************
	Create some triggers to enforce referential integrity
*********************************************************************
*/

create trigger checking_insert on checking for insert as
begin

if( (	select count(*) 
	from inserted i, customer c 
	where i.custid = c.custid  ) < 1 )
begin
	rollback tran
	print 'Checking Account Error: Must Be A Valid Customer To Have A Checking Account!'
end
end
go

create trigger checking_update on checking for insert as
begin

if update( custid )
begin
	rollback tran
	print 'Checking Account: You are not allowed modify the customer id!'
end
end
go


create trigger customer_delete on customer for delete as
begin

if( (	select count(*) 
	from deleted d, checking c 
	where d.custid = c.custid  ) > 0 )
begin
	rollback tran
	print 'Customer Error:  Unable To Delete A Customer With Active Checking Account'
end


if( (	select count(*) 
	from deleted d, savings c 
	where d.custid = c.custid  ) > 0 )
begin
	rollback tran
	print 'Customer Error:  Unable To Delete A Customer With Active Savings Account'
end
end
go


/*
*********************************************************************
	Create a rule to make sure accounts have more than 50.00
*********************************************************************
*/

create rule min_balance_rule as @balance > 50.00
go

sp_bindrule min_balance_rule, 'checking.balance'
go

sp_bindrule min_balance_rule, 'savings.balance'
go

create rule phone_rule as @phone like '([2-9][0-9][0-9]) [0-9][0-9][0-9]-[0-9][0-9][0-9][0-9]'
go

sp_bindrule phone_rule, 'customer.phone'
go

/*
*********************************************************************
	Create a default value for phone numbers
*********************************************************************
*/

create default phone_default as '(000) 000-0000'
go

sp_bindefault phone_default, 'customer.phone'
go

/*
*********************************************************************
	Add some sample data
*********************************************************************
*/

insert into customer(name,phone) values ('stewart','(206) 936-7374')
insert into customer(name,phone) values ('deby','(206) 936-4616')
insert into customer(name,phone) values ('gary','(206) 936-3615')
insert into customer(name,phone) values ('ron','(206) 936-4433')
insert into customer(name,phone) values ('jane','(206) 936-4216')
insert into customer(name,phone) values ('fred','(206) 936-1215')
go

insert into savings(custid,balance) values (1, 1000.00)
insert into savings(custid,balance) values (2, 2000.00)
insert into savings(custid,balance) values (3, 3000.00)
insert into savings(custid,balance) values (4, 4000.00)
insert into savings(custid,balance) values (5, 5000.00)
insert into savings(custid,balance) values (6, 6000.00)
go

insert into checking(custid,balance) values (1, 1000.00)
insert into checking(custid,balance) values (2, 2000.00)
insert into checking(custid,balance) values (3, 3000.00)
insert into checking(custid,balance) values (4, 4000.00)
insert into checking(custid,balance) values (5, 5000.00)
insert into checking(custid,balance) values (6, 6000.00)
go


/*
*********************************************************************
	Create a stored procedure
*********************************************************************
*/

create proc do_transfer @custid integer = 0, @amount decimal(10,2) = 0
as
begin

declare @err integer
declare @ctr integer

begin transaction

	update savings set balance = balance - @amount where custid = @custid

	select @err = @@error, @ctr = @@rowcount

	if( @err <> 0 )
	begin
		rollback tran
		return 1
	end

	if( @ctr <> 1 )
	begin
		rollback tran
		return 1
	end

	update checking set balance = balance + @amount where custid = @custid

	select @err = @@error

	if( @err <> 0 )
	begin
		rollback tran
		return 1
	end
	
	if( @ctr <> 1 )
	begin
		rollback tran
		return 1
	end


commit transaction


return 0

end
go


exec do_transfer 1, 100.00 
go

declare @ret_code integer

exec @ret_code = do_transfer 1, 100.00 

if(@ret_code = 0)
	select 'Success'
else
	select 'Failure'
go


declare @ret_code integer

exec @ret_code = do_transfer 100, 100.00 

if(@ret_code = 0)
	select 'Success'
else
	select 'Failure'
go

/*
***************************************************************
	drop all the tables and stored procedures
***************************************************************
*/

sp_unbindefault 'customer.phone'
go

drop default phone_default
go

sp_unbindrule 'checking.balance'
go

sp_unbindrule 'savings.balance'
go

sp_unbindrule 'customer.phone'
go

drop rule phone_rule
go

drop rule min_balance_rule
go

if exists (select * from sysobjects where name = 'do_transfer' )
begin
	drop proc do_transfer
end
go

if exists (select * from sysobjects where name = 'savings' and type = 'U')
begin
	drop table savings
end
go

if exists (select * from sysobjects where name = 'checking' and type = 'U')
begin
	drop table checking
end
go


if exists (select * from sysobjects where name = 'customer' and type = 'U')
begin
	drop table customer
end
go


